# -*- coding: utf-8 -*-
import pytest

from arkindex.exceptions import ErrorResponse


def test_request_no_more_retries(responses, dummy_client):
    # First call is an error
    responses.add(
        responses.POST,
        "https://dummy.test/api/v1/user/login/",
        status=502,
    )

    dummy_client.request.retry.wait = 0

    # There should be 5 retries
    max_retries = 5

    with pytest.raises(ErrorResponse):
        dummy_client.login("user@user.user", "Pa$$w0rd")

    assert len(responses.calls) == max_retries
    assert [call.request.url for call in responses.calls] == [
        "https://dummy.test/api/v1/user/login/",
    ] * max_retries



# -*- coding: utf-8 -*-

import os

import pytest

from arkindex.exceptions import ErrorResponse
from arkindex.mock import MockApiClient


def test_mock_client(responses):
    # Allow accessing remote API schemas, defaulting to the prod environment
    schema_url = os.environ.get(
        "ARKINDEX_API_SCHEMA_URL",
        "https://arkindex.teklia.com/api/v1/openapi/?format=json",
    )
    responses.add_passthru(schema_url)

    # Configure mock for dummy endpoints
    mock = MockApiClient()
    mock.add_response(
        "ListAnimals",
        [
            {
                "type": "cat",
                "name": "Henri",
            },
            {
                "type": "dog",
                "name": "Rick",
            },
        ],
    )
    invalid_body = mock.add_response(
        "CreateTag",
        name="Henri",
        body={"tag": "This is not the right payload"},
        response={
            "id": "1234-cat",
        },
    )
    mock.add_response(
        "CreateTag",
        name="Henri",
        body={"tag": "yyy"},
        response={
            "id": "1234-cat",
        },
    )
    extra_response = mock.add_response(
        "CreateTag",
        name="Nope",
        response={
            "id": "shouldNotAppear",
        },
    )
    mock.add_response(
        "CreateTag",
        name="Rick",
        body={"tag": "xxx"},
        response={
            "id": "5678-dog",
        },
    )
    mock.add_error_response(
        "DeleteTag",
        name="Henri",
        body={"tag": "yyy"},
        status_code=403,
        title="Forbidden",
        content="You cannot perform this action",
    )

    # Call the endpoints - this simulates business code
    animals = mock.paginate("ListAnimals")
    assert len(animals) == 2
    assert animals == [
        {
            "type": "cat",
            "name": "Henri",
        },
        {
            "type": "dog",
            "name": "Rick",
        },
    ]

    assert mock.request("CreateTag", name="Rick", body={"tag": "xxx"}) == {
        "id": "5678-dog",
    }
    assert mock.request("CreateTag", name="Henri", body={"tag": "yyy"}) == {
        "id": "1234-cat",
    }

    with pytest.raises(ErrorResponse) as e_info:
        mock.request("DeleteTag", name="Henri", body={"tag": "yyy"})
    error_response = e_info.value
    assert error_response.status_code == 403
    assert error_response.title == "Forbidden"
    assert error_response.content == "You cannot perform this action"

    # A request not mocked should raise an exception
    with pytest.raises(ErrorResponse):
        mock.request("DoSomething", parameter="value")

    # Check the mockup history
    assert len(mock.history) == 5
    assert [req.operation for req in mock.history] == [
        "ListAnimals",
        "CreateTag",
        "CreateTag",
        "DeleteTag",
        "DoSomething",
    ]

    # Almost all responses have been consumed
    assert len(mock.responses) == 2
    assert mock.responses == [invalid_body, extra_response]



# -*- coding: utf-8 -*-
from requests.auth import AuthBase


class TokenSessionAuthentication(AuthBase):

    safe_methods = ("GET", "HEAD", "OPTIONS", "TRACE")

    def __init__(
        self,
        token,
        scheme="Token",
        csrf_cookie_name="arkindex.csrf",
        csrf_header_name="X-CSRFToken",
    ):
        """
        :param str token: The API token to use.
        :param str scheme: The HTTP authentication scheme to use for token authentication.
        :param str csrf_cookie_name: Name of the CSRF token cookie.
        :param str csrf_header_name: Name of the CSRF request header.
        """
        self.token = token
        self.scheme = scheme
        self.csrf_cookie_name = csrf_cookie_name
        self.csrf_header_name = csrf_header_name
        self.csrf_token = None

    def store_csrf_token(self, response, **kwargs):
        if self.csrf_cookie_name in response.cookies:
            self.csrf_token = response.cookies[self.csrf_cookie_name]

    def __call__(self, request):
        # Add CSRF token
        if (
            self.csrf_token
            and self.csrf_header_name is not None
            and (request.method not in self.safe_methods)
        ):
            request.headers[self.csrf_header_name] = self.csrf_token

        if self.csrf_cookie_name is not None:
            request.register_hook("response", self.store_csrf_token)

        # Add API token
        if self.token is not None:
            request.headers["Authorization"] = f"{self.scheme} {self.token}"

        return request



# -*- coding: utf-8 -*-
import re
import typing

from arkindex.exceptions import SchemaError


class Document:
    def __init__(
        self,
        links: typing.Sequence["Link"],
        url: str = "",
    ):
        if not len(links):
            raise SchemaError(
                "An OpenAPI document must contain at least one valid operation."
            )

        links_by_name = {}

        # Ensure all names within a document are unique.
        for link in links:
            assert (
                link.name not in links_by_name
            ), f'Link "{link.name}" in Document must have a unique name.'
            links_by_name[link.name] = link

        self.links = links_by_name
        self.url = url


class Link:
    """
    Links represent the actions that a client may perform.
    """

    def __init__(
        self,
        url: str,
        method: str,
        handler: typing.Callable = None,
        name: str = "",
        encoding: str = "",
        fields: typing.Sequence["Field"] = None,
        deprecated: bool = False,
        paginated: typing.Optional[bool] = None,
    ):
        method = method.upper()
        fields = [] if (fields is None) else list(fields)

        url_path_names = set(
            [item.strip("{}").lstrip("+") for item in re.findall("{[^}]*}", url)]
        )
        path_fields = [field for field in fields if field.location == "path"]
        body_fields = [field for field in fields if field.location == "body"]

        assert method in (
            "GET",
            "POST",
            "PUT",
            "PATCH",
            "DELETE",
            "OPTIONS",
            "HEAD",
            "TRACE",
        )
        assert len(body_fields) < 2
        if body_fields:
            assert encoding
        for field in path_fields:
            assert field.name in url_path_names

        # Add in path fields for any "{param}" items that don't already have
        # a corresponding path field.
        for path_name in url_path_names:
            if path_name not in [field.name for field in path_fields]:
                fields += [Field(name=path_name, location="path", required=True)]

        self.url = url
        self.method = method
        self.handler = handler
        self.name = name if name else handler.__name__
        self.encoding = encoding
        self.fields = fields
        self.deprecated = deprecated
        self.paginated = paginated

    def get_path_fields(self):
        return [field for field in self.fields if field.location == "path"]

    def get_query_fields(self):
        return [field for field in self.fields if field.location == "query"]

    def get_body_field(self):
        for field in self.fields:
            if field.location == "body":
                return field
        return None

    def get_expanded_body(self):
        field = self.get_body_field()
        if field is None or not hasattr(field.schema, "properties"):
            return None
        return field.schema.properties


class Field:
    def __init__(
        self,
        name: str,
        location: str,
        required: bool = None,
        schema: typing.Any = None,
        example: typing.Any = None,
        deprecated: bool = False,
    ):
        assert location in ("path", "query", "body", "cookie", "header", "formData")
        if required is None:
            required = True if location in ("path", "body") else False
        if location == "path":
            assert required, "May not set 'required=False' on path fields."

        self.name = name
        self.location = location
        self.required = required
        self.schema = schema
        self.example = example
        self.deprecated = deprecated



# -*- coding: utf-8 -*-
from pathlib import Path

import pytest
import responses

from arkindex import ArkindexClient
from arkindex.exceptions import SchemaError

DUMMY_SCHEMA = Path(__file__).absolute().parent / "schema.json"


def test_invalid_url():
    with pytest.raises(
        SchemaError,
        match=r"Could not retrieve a proper OpenAPI schema from http://aaa/api/v1/openapi/\?format=json",
    ):
        ArkindexClient(base_url="http://aaa")


def test_http_error():
    responses.add(
        responses.GET,
        "https://dummy.test/api/v1/openapi/?format=json",
        status=418,
    )
    with pytest.raises(SchemaError):
        ArkindexClient(base_url="https://dummy.test/")


def test_invalid_json():
    responses.add(
        responses.GET,
        "https://dummy.test/api/v1/openapi/?format=json",
        body="{",
    )
    with pytest.raises(SchemaError):
        ArkindexClient(base_url="https://dummy.test/")


def test_no_endpoints():
    responses.add(
        responses.GET,
        "https://dummy.test/api/v1/openapi/?format=json",
        json={
            "openapi": "3.0.2",
            "info": {"title": "/dev/null", "version": "0"},
            "paths": {},
        },
    )
    with pytest.raises(
        SchemaError,
        match="An OpenAPI document must contain at least one valid operation.",
    ):
        ArkindexClient(base_url="https://dummy.test/")


def test_schema_url():
    with DUMMY_SCHEMA.open() as f:
        responses.add(
            responses.GET,
            "http://dev/null",
            body=f.read(),
        )
    cli = ArkindexClient(base_url="https://dummy.test/", schema_url="http://dev/null")
    assert cli.document.url == "https://dummy.test/"
    assert list(cli.document.links.keys()) == [
        "ListElements",
        "ListProcessElements",
        "UploadDataFile",
        "ListElementMetaData",
        "Login",
    ]


def test_local_path():
    cli = ArkindexClient(base_url="https://dummy.test/", schema_url=str(DUMMY_SCHEMA))
    assert cli.document.url == "https://dummy.test/"
    assert list(cli.document.links.keys()) == [
        "ListElements",
        "ListProcessElements",
        "UploadDataFile",
        "ListElementMetaData",
        "Login",
    ]



# -*- coding: utf-8 -*-
import typing

import typesystem

from arkindex.schema.openapi import OPEN_API, OpenAPI


def validate(schema: typing.Union[dict, str, bytes]):
    if not isinstance(schema, (dict, str, bytes)):
        raise ValueError("schema must be either str, bytes, or dict.")

    if isinstance(schema, bytes):
        schema = schema.decode("utf8", "ignore")

    if isinstance(schema, str):
        token = typesystem.tokenize_json(schema)
        value = typesystem.validate_with_positions(token=token, validator=OpenAPI)
    else:
        value = OPEN_API.validate(schema)

    return OpenAPI().load(value)



# -*- coding: utf-8 -*-
from collections import namedtuple

Position = namedtuple("Position", ["line_no", "column_no", "index"])


class ErrorMessage:
    def __init__(self, text, code, index=None, position=None):
        self.text = text
        self.code = code
        self.index = index
        self.position = position

    def __eq__(self, other):
        return (
            self.text == other.text
            and self.code == other.code
            and self.index == other.index
            and self.position == other.position
        )

    def __repr__(self):
        return "%s(%s, code=%s, index=%s, position=%s)" % (
            self.__class__.__name__,
            repr(self.text),
            repr(self.code),
            repr(self.index),
            repr(self.position),
        )


class ErrorResponse(Exception):
    """
    Raised when a client request results in an error response being returned.
    """

    def __init__(self, title, status_code, content):
        self.title = title
        self.status_code = status_code
        self.content = content

    def __repr__(self):
        return "%s(%s, status_code=%s, content=%s)" % (
            self.__class__.__name__,
            repr(self.title),
            repr(self.status_code),
            repr(self.content),
        )

    def __str__(self):
        return repr(self.content)


class ClientError(Exception):
    """
    Raised when a client is unable to fulfil an API request.
    """

    def __init__(self, messages):
        self.messages = messages
        super().__init__(messages)

    def __repr__(self):
        return "%s(messages=%s)" % (
            self.__class__.__name__,
            repr(self.messages),
        )

    def __str__(self):
        if len(self.messages) == 1 and not self.messages[0].index:
            return self.messages[0].text
        return str(self.messages)


class SchemaError(Exception):
    """
    Any error occurring during the acquisition and processing of the OpenAPI schema.
    """



# -*- coding: utf-8 -*-
"""
Arkindex API Client
"""
import json
import logging
import os
import warnings
from importlib.metadata import version
from time import sleep
from urllib.parse import quote, urljoin, urlparse, urlsplit

import requests
import typesystem
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_exponential,
)

from arkindex.auth import TokenSessionAuthentication
from arkindex.client import decoders
from arkindex.exceptions import ClientError, ErrorMessage, ErrorResponse, SchemaError
from arkindex.pagination import ResponsePaginator
from arkindex.schema.validator import validate

logger = logging.getLogger(__name__)

REQUEST_TIMEOUT = (30, 60)

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

DEFAULT_BASE_URL = "https://arkindex.teklia.com/"

# Endpoint accessed by the client on instantiation to retrieve the OpenAPI schema
SCHEMA_ENDPOINT = "/api/v1/openapi/?format=json"


def _is_500_error(exc: Exception) -> bool:
    """
    Check if an Arkindex API error is a 50x
    This is used to retry most API calls implemented here
    """
    if not isinstance(exc, ErrorResponse):
        return False

    return 500 <= exc.status_code < 600


def options_from_env():
    """
    Get API client keyword arguments from environment variables.
    """
    options = {}
    # ARKINDEX_TASK_TOKEN takes priority over ARKINDEX_API_TOKEN
    if "ARKINDEX_TASK_TOKEN" in os.environ:
        options["auth_scheme"] = "Ponos"
        options["token"] = os.environ.get("ARKINDEX_TASK_TOKEN")
    elif "ARKINDEX_API_TOKEN" in os.environ:
        options["auth_scheme"] = "Token"
        options["token"] = os.environ.get("ARKINDEX_API_TOKEN")

    # Allow overriding the default auth schemes
    if "ARKINDEX_API_AUTH_SCHEME" in os.environ:
        options["auth_scheme"] = os.environ.get("ARKINDEX_API_AUTH_SCHEME")

    if "ARKINDEX_API_URL" in os.environ:
        options["base_url"] = os.environ.get("ARKINDEX_API_URL")

    if "ARKINDEX_API_SCHEMA_URL" in os.environ:
        options["schema_url"] = os.environ.get("ARKINDEX_API_SCHEMA_URL")

    if "ARKINDEX_API_CSRF_COOKIE" in os.environ:
        options["csrf_cookie"] = os.environ.get("ARKINDEX_API_CSRF_COOKIE")

    return options


class ArkindexClient:
    """
    An Arkindex API client.
    """

    def __init__(
        self,
        token=None,
        auth_scheme="Token",
        base_url=DEFAULT_BASE_URL,
        schema_url=None,
        csrf_cookie=None,
        sleep=0,
        verify=True,
    ):
        r"""
        :param token: An API token to use. If omitted, access is restricted to public endpoints.
        :type token: str or None
        :param str base_url: A custom base URL for the client. If omitted, defaults to the Arkindex main server.
        :param schema_url: URL or local path to an OpenAPI schema to use instead of the Arkindex instance's own schema.
        :type schema_url: str or None
        :param csrf_cookie: Use a custom CSRF cookie name. By default, the client will try to use any cookie
           defined in ``x-csrf-cookie`` on the Server Object of the OpenAPI specification, and fall back to
           ``arkindex.csrf``.
        :type csrf_cookie: str or None
        :param float sleep: Number of seconds to wait before sending each API request,
           as a simple means of throttling.
        :param bool verify: Whether to verify the SSL certificate on each request. Enabled by default.
        """
        self.decoders = [
            decoders.JSONDecoder(),
            decoders.TextDecoder(),
            decoders.DownloadDecoder(),
        ]

        self.session = requests.Session()
        self.session.verify = verify
        client_version = version("arkindex-client")
        self.session.headers.update(
            {
                "accept": ", ".join([decoder.media_type for decoder in self.decoders]),
                "user-agent": f"arkindex-client/{client_version}",
            }
        )

        if not schema_url:
            schema_url = urljoin(base_url, SCHEMA_ENDPOINT)

        try:
            split = urlsplit(schema_url)
            if split.scheme == "file" or not (split.scheme or split.netloc):
                # This is a local path
                with open(schema_url) as f:
                    schema = json.load(f)
            else:
                resp = self.session.get(
                    schema_url,
                    headers={
                        # Explicitly request an OpenAPI schema in JSON and not YAML
                        "Accept": "application/vnd.oai.openapi+json, application/json",
                    },
                )
                resp.raise_for_status()
                schema = resp.json()
        except Exception as e:
            raise SchemaError(
                f"Could not retrieve a proper OpenAPI schema from {schema_url}"
            ) from e

        self.document = validate(schema)

        # Try to autodetect the CSRF cookie:
        # - Try to find a matching server for this base URL and look for the x-csrf-cookie extension
        # - Fallback to arkindex.csrf
        if not csrf_cookie:
            split_base_url = urlsplit(base_url or self.document.url)
            csrf_cookies = {
                urlsplit(server.get("url", "")).netloc: server.get("x-csrf-cookie")
                for server in schema.get("servers", [])
            }
            csrf_cookie = csrf_cookies.get(split_base_url.netloc) or "arkindex.csrf"

        self.configure(
            token=token,
            auth_scheme=auth_scheme,
            base_url=base_url,
            csrf_cookie=csrf_cookie,
            sleep=sleep,
        )

    def __repr__(self):
        return "<{} on {!r}>".format(
            self.__class__.__name__,
            self.document.url if hasattr(self, "document") else "",
        )

    def configure(
        self,
        token=None,
        auth_scheme="Token",
        base_url=None,
        csrf_cookie=None,
        sleep=None,
    ):
        """
        Reconfigure the API client.

        :param token: An API token to use. If omitted, access is restricted to public endpoints.
        :type token: str or None
        :param auth_scheme:
           An authentication scheme to use. This is added in the HTTP header before the token:
           ``Authorization: [scheme] [token]``.
           This should use ``Token`` to authenticate as a regular user and ``Ponos`` to authenticate as a Ponos task.
           If omitted, this defaults to ``Token``.
        :type auth_scheme: str or None
        :param base_url: A custom base URL for the client. If omitted, defaults to the Arkindex main server.
        :type base_url: str or None
        :param csrf_cookie: Use a custom CSRF cookie name. Falls back to ``arkindex.csrf``.
        :type csrf_cookie: str or None
        :param float sleep: Number of seconds to wait before sending each API request,
           as a simple means of throttling.
        """
        if not csrf_cookie:
            csrf_cookie = "arkindex.csrf"
        self.session.auth = TokenSessionAuthentication(
            token,
            csrf_cookie_name=csrf_cookie,
            scheme=auth_scheme,
        )

        if not sleep or not isinstance(sleep, float) or sleep < 0:
            self.sleep_duration = 0
        self.sleep_duration = sleep

        if base_url:
            self.document.url = base_url

        # Add the Referer header to allow Django CSRF to function
        self.session.headers.setdefault("Referer", self.document.url)

    def lookup_operation(self, operation_id: str):
        if operation_id in self.document.links:
            return self.document.links[operation_id]

        text = 'Operation ID "%s" not found in schema.' % operation_id
        message = ErrorMessage(text=text, code="invalid-operation")
        raise ClientError(messages=[message])

    def paginate(self, operation_id, *args, **kwargs):
        """
        Perform a usual API request, but handle paginated endpoints.

        :return: An iterator for a paginated endpoint.
        :rtype: Union[arkindex.pagination.ResponsePaginator, dict, list]
        """

        link = self.lookup_operation(operation_id)
        # If there was no x-paginated, trust the caller and assume the endpoint is paginated
        if link.paginated is not False:
            return ResponsePaginator(self, operation_id, *args, **kwargs)
        return self.request(operation_id, *args, **kwargs)

    def login(self, email, password):
        """
        Login to Arkindex using an email/password combination.
        This helper method automatically sets the client's authentication settings with the token.
        """
        resp = self.request("Login", body={"email": email, "password": password})
        if "auth_token" in resp:
            self.session.auth.scheme = "Token"
            self.session.auth.token = resp["auth_token"]
        return resp

    def get_query_params(self, link, params):
        return {
            field.name: params[field.name]
            for field in link.get_query_fields()
            if field.name in params
        }

    def get_url(self, link, params):
        url = urljoin(self.document.url, link.url)

        scheme = urlparse(url).scheme.lower()

        if not scheme:
            text = "URL missing scheme '%s'." % url
            message = ErrorMessage(text=text, code="invalid-url")
            raise ClientError(messages=[message])

        if scheme not in ("http", "https"):
            text = "Unsupported URL scheme '%s'." % scheme
            message = ErrorMessage(text=text, code="invalid-url")
            raise ClientError(messages=[message])

        for field in link.get_path_fields():
            value = str(params[field.name])
            if "{%s}" % field.name in url:
                url = url.replace("{%s}" % field.name, quote(value, safe=""))
            elif "{+%s}" % field.name in url:
                url = url.replace("{+%s}" % field.name, quote(value, safe="/"))

        return url

    def get_content(self, link, params):
        body_field = link.get_body_field()
        if body_field and body_field.name in params:
            assert (
                link.encoding == "application/json"
            ), "Only JSON request bodies are supported"
            return params[body_field.name]

    def get_decoder(self, content_type=None):
        """
        Given the value of a 'Content-Type' header, return the appropriate
        decoder for handling the response content.
        """
        if content_type is None:
            return self.decoders[0]

        content_type = content_type.split(";")[0].strip().lower()
        main_type = content_type.split("/")[0] + "/*"
        wildcard_type = "*/*"

        for codec in self.decoders:
            if codec.media_type in (content_type, main_type, wildcard_type):
                return codec

        text = (
            "Unsupported encoding '%s' in response Content-Type header." % content_type
        )
        message = ErrorMessage(text=text, code="cannot-decode-response")
        raise ClientError(messages=[message])

    def single_request(self, operation_id, **parameters):
        """
        Perform an API request.

        :param str operation_id: Name of the API endpoint.
        :param path_parameters: Path parameters for this endpoint.
        """
        link = self.lookup_operation(operation_id)
        if link.deprecated:
            warnings.warn(
                "Endpoint '{}' is deprecated.".format(operation_id),
                DeprecationWarning,
                stacklevel=2,
            )

        validator = typesystem.Object(
            properties={field.name: typesystem.Any() for field in link.fields},
            required=[field.name for field in link.fields if field.required],
            additional_properties=False,
        )
        try:
            validator.validate(parameters)
        except typesystem.ValidationError as exc:
            raise ClientError(messages=exc.messages()) from None

        method = link.method
        url = self.get_url(link, parameters)

        content = self.get_content(link, parameters)
        query_params = self.get_query_params(link, parameters)
        fields = link.get_query_fields()

        for field in fields:
            if field.deprecated and field.name in query_params:
                warnings.warn(
                    "Parameter '{}' is deprecated.".format(field.name),
                    DeprecationWarning,
                    stacklevel=2,
                )

        if self.sleep_duration:
            logger.debug(
                "Delaying request by {:f} seconds...".format(self.sleep_duration)
            )
            sleep(self.sleep_duration)

        return self._send_request(
            method, url, query_params=query_params, content=content
        )

    def _send_request(self, method, url, query_params=None, content=None):
        options = {
            "params": query_params,
            "timeout": REQUEST_TIMEOUT,
        }
        if content is not None:
            options["json"] = content

        response = self.session.request(method, url, **options)

        # Given an HTTP response, return the decoded data.
        result = None
        if response.content:
            content_type = response.headers.get("content-type")
            decoder = self.get_decoder(content_type)
            result = decoder.decode(response)

        if 400 <= response.status_code <= 599:
            title = "%d %s" % (response.status_code, response.reason)
            raise ErrorResponse(
                title=title, status_code=response.status_code, content=result
            )

        return result

    @retry(
        retry=retry_if_exception(_is_500_error),
        wait=wait_exponential(multiplier=2, min=3),
        reraise=True,
        stop=stop_after_attempt(5),
        before_sleep=before_sleep_log(logger, logging.INFO),
    )
    def request(self, operation_id, **parameters):
        """
        Perform an API request with an automatic retry mechanism in case of 50X errors.
        A failing API call will be retried 5 times, with an exponential sleep time going
        through 3, 4, 8 and 16 seconds of wait between call.
        If the 5th call still gives a 50x, the exception is re-raised and the caller should catch it.
        Log messages are displayed before sleeping (when at least one exception occurred).

        :param str operation_id: Name of the API endpoint.
        :param parameters: Body, Path or Query parameters passed as kwargs.
            Body parameters must be passed using the `body` keyword argument, others can be set directly.

        Example usage for POST and unpaginated GET requests:

        >>> request(
        ...     "CreateMetaDataBulk",
        ...     id="8f8f196f-49bc-444e-9cfe-c705c3cd01ae",
        ...     body={
        ...         "worker_run_id": "50e1f2d4-2087-41ed-a862-d17576bae480",
        ...         "metadata_list": [
        ...             …
        ...         ],
        ...     },
        ... )
        >>> request(
        ...     "ListElements",
        ...     corpus="7358ab03-cc36-4160-86ce-98f70e993a0f",
        ...     top_level=True,
        ... )
        """
        return self.single_request(operation_id, **parameters)



# -*- coding: utf-8 -*-
from responses import matchers


def test_login(responses, dummy_client):
    responses.add(
        responses.POST,
        "https://dummy.test/api/v1/user/login/",
        json={"auth_token": "tolkien"},
    )

    assert dummy_client.session.auth.token is None
    dummy_client.session.auth.scheme = "Beer"

    assert dummy_client.login("user@user.user", "Pa$$w0rd") == {"auth_token": "tolkien"}

    assert dummy_client.session.auth.token == "tolkien"
    assert dummy_client.session.auth.scheme == "Token"


def test_default_auth_scheme(responses, dummy_client):
    responses.add(
        responses.GET,
        "https://dummy.test/api/v1/elements/",
        match=[matchers.header_matcher({"Authorization": "Token ofgratitude"})],
        json={"count": 0, "previous": None, "next": None, "results": []},
    )

    dummy_client.configure(token="ofgratitude")

    assert dummy_client.request("ListElements") == {
        "count": 0,
        "previous": None,
        "next": None,
        "results": [],
    }


def test_custom_auth_scheme(responses, dummy_client):
    responses.add(
        responses.GET,
        "https://dummy.test/api/v1/elements/",
        match=[matchers.header_matcher({"Authorization": "Digest Food"})],
        json={"count": 0, "previous": None, "next": None, "results": []},
    )

    dummy_client.configure(token="Food", auth_scheme="Digest")

    assert dummy_client.request("ListElements") == {
        "count": 0,
        "previous": None,
        "next": None,
        "results": [],
    }



# -*- coding: utf-8 -*-
import json
import os
import posixpath
import shutil
import tempfile
from email.message import EmailMessage
from urllib.parse import urlparse

from arkindex.compat import DownloadedFile


class BaseDecoder:
    media_type = None

    def decode(self, bytestring, **options):
        raise NotImplementedError()


class JSONDecoder(BaseDecoder):
    media_type = "application/json"

    def decode(self, response):
        """
        Return raw JSON data.
        """
        content = response.content.decode("utf-8")
        return json.loads(content)


class TextDecoder(BaseDecoder):
    media_type = "text/*"

    def decode(self, response):
        return response.text


class DownloadDecoder(BaseDecoder):
    """
    A codec to handle raw file downloads, such as images and other media.
    """

    media_type = "*/*"

    def __init__(self, download_dir=None):
        """
        `download_dir` - If `None` then downloaded files will be temporary files
        that are deleted on close. If set to a value, then downloaded files
        will be saved to this directory, and will not be automatically deleted.
        """
        self._delete_on_close = download_dir is None
        self.download_dir = download_dir

    def decode(self, response):
        base_url = response.url
        content_type = response.headers.get("content-type")
        content_disposition = response.headers.get("content-disposition")

        # Write the download to a temporary .download file.
        fd, temp_path = tempfile.mkstemp(suffix=".download")
        with os.fdopen(fd, "wb") as file_handle:
            for chunk in response.iter_content(chunk_size=4096):
                file_handle.write(chunk)

        # Determine the output filename.
        output_filename = _get_filename(base_url, content_type, content_disposition)

        # Determine the output directory.
        output_dir = self.download_dir
        if output_dir is None:
            output_dir = os.path.dirname(temp_path)

        # Determine the full output path.
        output_path = os.path.join(output_dir, output_filename)

        # Move the temporary download file to the final location.
        if output_path != temp_path:
            output_path = _unique_output_path(output_path)
            shutil.move(temp_path, output_path)

        # Open the file and return the file object.
        output_file = open(output_path, "rb")
        downloaded = DownloadedFile(
            output_file, output_path, delete=self._delete_on_close
        )
        downloaded.basename = output_filename
        return downloaded


def _guess_extension(content_type):
    """
    Python's `mimetypes.guess_extension` is no use because it simply returns
    the first of an unordered set. We use the same set of media types here,
    but take a reasonable preference on what extension to map to.
    """
    return {
        "application/javascript": ".js",
        "application/msword": ".doc",
        "application/octet-stream": ".bin",
        "application/oda": ".oda",
        "application/pdf": ".pdf",
        "application/pkcs7-mime": ".p7c",
        "application/postscript": ".ps",
        "application/vnd.apple.mpegurl": ".m3u",
        "application/vnd.ms-excel": ".xls",
        "application/vnd.ms-powerpoint": ".ppt",
        "application/x-bcpio": ".bcpio",
        "application/x-cpio": ".cpio",
        "application/x-csh": ".csh",
        "application/x-dvi": ".dvi",
        "application/x-gtar": ".gtar",
        "application/x-hdf": ".hdf",
        "application/x-latex": ".latex",
        "application/x-mif": ".mif",
        "application/x-netcdf": ".nc",
        "application/x-pkcs12": ".p12",
        "application/x-pn-realaudio": ".ram",
        "application/x-python-code": ".pyc",
        "application/x-sh": ".sh",
        "application/x-shar": ".shar",
        "application/x-shockwave-flash": ".swf",
        "application/x-sv4cpio": ".sv4cpio",
        "application/x-sv4crc": ".sv4crc",
        "application/x-tar": ".tar",
        "application/x-tcl": ".tcl",
        "application/x-tex": ".tex",
        "application/x-texinfo": ".texinfo",
        "application/x-troff": ".tr",
        "application/x-troff-man": ".man",
        "application/x-troff-me": ".me",
        "application/x-troff-ms": ".ms",
        "application/x-ustar": ".ustar",
        "application/x-wais-source": ".src",
        "application/xml": ".xml",
        "application/zip": ".zip",
        "audio/basic": ".au",
        "audio/mpeg": ".mp3",
        "audio/x-aiff": ".aif",
        "audio/x-pn-realaudio": ".ra",
        "audio/x-wav": ".wav",
        "image/gif": ".gif",
        "image/ief": ".ief",
        "image/jpeg": ".jpe",
        "image/png": ".png",
        "image/svg+xml": ".svg",
        "image/tiff": ".tiff",
        "image/vnd.microsoft.icon": ".ico",
        "image/x-cmu-raster": ".ras",
        "image/x-ms-bmp": ".bmp",
        "image/x-portable-anymap": ".pnm",
        "image/x-portable-bitmap": ".pbm",
        "image/x-portable-graymap": ".pgm",
        "image/x-portable-pixmap": ".ppm",
        "image/x-rgb": ".rgb",
        "image/x-xbitmap": ".xbm",
        "image/x-xpixmap": ".xpm",
        "image/x-xwindowdump": ".xwd",
        "message/rfc822": ".eml",
        "text/css": ".css",
        "text/csv": ".csv",
        "text/html": ".html",
        "text/plain": ".txt",
        "text/richtext": ".rtx",
        "text/tab-separated-values": ".tsv",
        "text/x-python": ".py",
        "text/x-setext": ".etx",
        "text/x-sgml": ".sgml",
        "text/x-vcard": ".vcf",
        "text/xml": ".xml",
        "video/mp4": ".mp4",
        "video/mpeg": ".mpeg",
        "video/quicktime": ".mov",
        "video/webm": ".webm",
        "video/x-msvideo": ".avi",
        "video/x-sgi-movie": ".movie",
    }.get(content_type, "")


def _unique_output_path(path):
    """
    Given a path like '/a/b/c.txt'

    Return the first available filename that doesn't already exist,
    using an incrementing suffix if needed.

    For example: '/a/b/c.txt' or '/a/b/c (1).txt' or '/a/b/c (2).txt'...
    """
    basename, ext = os.path.splitext(path)
    idx = 0
    while os.path.exists(path):
        idx += 1
        path = "%s (%d)%s" % (basename, idx, ext)
    return path


def _safe_filename(filename):
    """
    Sanitize output filenames, to remove any potentially unsafe characters.
    """
    filename = os.path.basename(filename)

    keepcharacters = (" ", ".", "_", "-")
    filename = (
        "".join(char for char in filename if char.isalnum() or char in keepcharacters)
        .strip()
        .strip(".")
    )

    return filename


def _get_filename_from_content_disposition(content_disposition):
    """
    Determine an output filename based on the `Content-Disposition` header.
    """
    message = EmailMessage()
    message["content-disposition"] = content_disposition
    filename = message["content-disposition"].params.get("filename")

    if filename:
        return _safe_filename(filename)

    return None


def _get_filename_from_url(url, content_type=None):
    """
    Determine an output filename based on the download URL.
    """
    parsed = urlparse(url)
    final_path_component = posixpath.basename(parsed.path.rstrip("/"))
    filename = _safe_filename(final_path_component)
    suffix = _guess_extension(content_type or "")

    if filename:
        if "." not in filename:
            return filename + suffix
        return filename

    return "download" + suffix


def _get_filename(base_url, content_type=None, content_disposition=None):
    """
    Determine an output filename to use for the download.
    """
    filename = None
    if content_disposition:
        filename = _get_filename_from_content_disposition(content_disposition)
        if filename is not None:
            return filename

    return _get_filename_from_url(base_url, content_type)



# -*G- coding: utf-8 -*-
import logging
import time

import pytest
import responses
from responses import matchers

from arkindex import ArkindexClient
from arkindex.pagination import PaginationMode


def test_pagination_empty(mock_schema):
    responses.add(
        responses.GET,
        "https://dummy.test/api/v1/elements/",
        json={"count": 0, "previous": None, "next": None, "results": []},
    )

    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    with pytest.raises(StopIteration):
        next(cli.paginate("ListElements"))


def test_page_mode_detection(mock_schema):
    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    pagination = cli.paginate("ListElements")
    responses.assert_call_count(
        count=1, url="https://dummy.test/api/v1/openapi/?format=json"
    )
    assert pagination.mode is PaginationMode.PageNumber


def test_cursor_mode_detection(mock_schema):
    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    pagination = cli.paginate("ListProcessElements", id="process_id")
    responses.assert_call_count(
        count=1, url="https://dummy.test/api/v1/openapi/?format=json"
    )
    assert pagination.mode is PaginationMode.Cursor


def test_page_pagination_with_missing_data(mock_schema, monkeypatch):
    """
    Page number pagination skips erroneous pages when allow_missing_data parameter is set
    """
    base_url = "https://dummy.test/api/v1/elements/"

    # Disable sleeps
    monkeypatch.setattr(time, "sleep", lambda x: None)

    # Page 1
    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({})],
        json={
            "count": 9,
            "number": 1,
            "previous": None,
            "next": f"{base_url}?page=2",
            "results": [1, 2, 3],
        },
    )

    # Page 2 is erroneous
    # We need 3 responses, one for each retries
    for i in range(3):
        responses.add(
            responses.GET,
            base_url,
            match=[matchers.query_param_matcher({"page": 2})],
            status=400,
            json={"error": "some error happened"},
        )

    # Page 3
    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({"page": 3})],
        json={
            "count": 9,
            "number": 3,
            "previous": f"{base_url}?page=3",
            "next": None,
            "results": [7, 8, 9],
        },
    )

    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    paginator = cli.paginate("ListElements", allow_missing_data=True, retries=3)
    assert paginator.allow_missing_data is True

    # First page is retrieved
    assert len(paginator) == 9
    assert len(responses.calls) == 2
    responses.assert_call_count(count=1, url=base_url)
    # Remaining pages are retrieved with retries
    assert list(paginator) == [1, 2, 3, 7, 8, 9]
    assert len(responses.calls) == 6
    responses.assert_call_count(count=3, url=f"{base_url}?page=2")
    # Page 2 has failed
    assert paginator.missing == {2}


def test_page_pagination_incomplete(mock_schema, monkeypatch):
    """
    Pagination raises an exception on repeated errors by default
    """
    base_url = "https://dummy.test/api/v1/elements/"

    # Disable sleeps
    monkeypatch.setattr(time, "sleep", lambda x: None)

    # Page 1
    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({})],
        json={
            "count": 9,
            "number": 1,
            "previous": None,
            "next": f"{base_url}?page=2",
            "results": ["A", "B"],
        },
    )

    # Page 2 is erroneous
    # Page 3 is not needed as it won't try to load it
    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({"page": 2})],
        status=500,
    )

    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    paginator = cli.paginate("ListElements", retries=1)
    assert paginator.allow_missing_data is False

    with pytest.raises(
        Exception, match="Stopping pagination as data will be incomplete"
    ):
        list(paginator)


def test_cursor_pagination_length(mock_schema):
    """
    The first page should be retrieved with the `with_count` parameter in case of a cursor pagination
    Allows the client to know the total number of results
    """
    base_url = "https://dummy.test/api/v1/process/process_id/elements/"

    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({"with_count": "true"})],
        json={
            "count": 6,
            "previous": None,
            "next": f"{base_url}?cursor=DEF",
            "results": [1, 2, 3],
        },
    )
    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({"cursor": "DEF"})],
        json={
            "count": None,
            "previous": f"{base_url}?cursor=ABC",
            "next": None,
            "results": [4, 5, 6],
        },
    )

    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    paginator = cli.paginate("ListProcessElements", id="process_id")

    assert len(paginator) == 6
    assert len(responses.calls) == 2
    # Schema is used to detect the pagination mode
    responses.assert_call_count(
        count=1, url="https://dummy.test/api/v1/openapi/?format=json"
    )
    responses.assert_call_count(count=1, url=f"{base_url}?with_count=true")
    assert list(paginator) == list(range(1, 7))
    assert len(responses.calls) == 3


def test_cursor_pagination_zero_results(mock_schema):
    """
    Pagination should handle queries returning zero results
    """
    base_url = "https://dummy.test/api/v1/process/process_id/elements/"
    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({"with_count": "true"})],
        json={
            "count": 0,
            "previous": None,
            "next": None,
            "results": [],
        },
    )

    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    paginator = cli.paginate("ListProcessElements", id="process_id")

    assert len(paginator) == 0
    assert list(paginator) == []
    assert len(responses.calls) == 2


def test_cursor_pagination_len_errors(mock_schema, mocker):
    """
    Errors must be raised when calling the __len__ method of an iterator
    """
    base_url = "https://dummy.test/api/v1/process/process_id/elements/"
    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({"with_count": "true"})],
        status=500,
    )

    # Prevent waiting for a random amount of time between retries
    mocker.patch("arkindex.pagination.random.random", return_value=0.0001)

    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    paginator = cli.paginate("ListProcessElements", id="process_id", retries=5)

    with pytest.raises(
        Exception, match="Stopping pagination as data will be incomplete"
    ):
        len(paginator)
    assert len(responses.calls) == 6


def test_paginate_x_paginated(mock_schema):
    responses.add(
        responses.GET,
        "https://dummy.test/api/v1/element/element_id/metadata/",
        match=[matchers.query_param_matcher({})],
        json=["a", "b"],
    )

    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    assert cli.paginate("ListElementMetaData", id="element_id") == ["a", "b"]
    assert len(responses.calls) == 2


def test_paginate_from_initial_page(mock_schema, caplog):
    caplog.set_level(logging.INFO)
    base_url = "https://dummy.test/api/v1/elements/"

    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({"page": "2"})],
        json={
            "count": 9,
            "number": 1,
            "previous": f"{base_url}?page=1",
            "next": f"{base_url}?page=3",
            "results": [4, 5, 6],
        },
    )
    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({"page": "3"})],
        json={
            "count": 9,
            "number": 1,
            "previous": f"{base_url}?page=2",
            "next": None,
            "results": [7, 8, 9],
        },
    )

    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    assert list(cli.paginate("ListElements", page=2)) == [4, 5, 6, 7, 8, 9]
    assert len(responses.calls) == 3
    assert [(level, message) for _module, level, message in caplog.record_tuples] == [
        (logging.INFO, "Loading page 2 on try 1/5"),
        (logging.INFO, "Pagination will load a total of 2 pages."),
        (logging.INFO, "Loading page 3 on try 1/5 - remains 1 page to load."),
    ]


def test_paginate_from_initial_cursor(mock_schema, caplog):
    caplog.set_level(logging.INFO)
    base_url = "https://dummy.test/api/v1/process/process_id/elements/"

    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({"cursor": "DEF", "with_count": "true"})],
        json={
            "count": 7,
            "previous": f"{base_url}?cursor=ABC",
            "next": f"{base_url}?cursor=GHI",
            "results": [4, 5, 6],
        },
    )
    responses.add(
        responses.GET,
        base_url,
        match=[matchers.query_param_matcher({"cursor": "GHI"})],
        json={
            "count": 7,
            "previous": f"{base_url}?cursor=ABC",
            "next": None,
            "results": [7],
        },
    )
    cli = ArkindexClient("t0k3n", base_url="https://dummy.test")
    assert list(cli.paginate("ListProcessElements", id="process_id", cursor="DEF")) == [
        4,
        5,
        6,
        7,
    ]
    assert len(responses.calls) == 3
    assert [(level, message) for _module, level, message in caplog.record_tuples] == [
        (logging.INFO, "Loading page DEF on try 1/5"),
        (logging.INFO, "Pagination will load a maximum of 3 pages"),
        (
            logging.INFO,
            "Loading cursor GHI on try 1/5 - remains a maximum of 2 pages to load.",
        ),
    ]



# -*- coding: utf-8 -*-
import os

import pytest

from arkindex import options_from_env


@pytest.mark.parametrize(
    "env,expected",
    [
        ({}, {}),
        ({"ARKINDEX_API_URL": "http://lol"}, {"base_url": "http://lol"}),
        ({"ARKINDEX_API_SCHEMA_URL": "http://lol"}, {"schema_url": "http://lol"}),
        ({"ARKINDEX_API_AUTH_SCHEME": "Bribe"}, {"auth_scheme": "Bribe"}),
        (
            {"ARKINDEX_API_CSRF_COOKIE": "raisin-oatmeal"},
            {"csrf_cookie": "raisin-oatmeal"},
        ),
        (
            {"ARKINDEX_API_TOKEN": "ofGoodWill"},
            {"auth_scheme": "Token", "token": "ofGoodWill"},
        ),
        (
            {"ARKINDEX_TASK_TOKEN": "Kimonos"},
            {"auth_scheme": "Ponos", "token": "Kimonos"},
        ),
        # The task token should override the API token
        (
            {"ARKINDEX_API_TOKEN": "ofGoodWill", "ARKINDEX_TASK_TOKEN": "Kimonos"},
            {"auth_scheme": "Ponos", "token": "Kimonos"},
        ),
        # The auth scheme can override the scheme set by the API token
        (
            {"ARKINDEX_API_TOKEN": "Bad", "ARKINDEX_API_AUTH_SCHEME": "Breaking"},
            {"auth_scheme": "Breaking", "token": "Bad"},
        ),
        # The auth scheme can override the scheme set by the task token
        (
            {"ARKINDEX_TASK_TOKEN": "Bad", "ARKINDEX_API_AUTH_SCHEME": "Breaking"},
            {"auth_scheme": "Breaking", "token": "Bad"},
        ),
        # Also try with both tokens set
        (
            {
                "ARKINDEX_API_AUTH_SCHEME": "It",
                "ARKINDEX_API_TOKEN": "Fails",
                "ARKINDEX_TASK_TOKEN": "Works",
            },
            {"auth_scheme": "It", "token": "Works"},
        ),
        # Set everything at once
        (
            {
                "ARKINDEX_API_URL": "http://lol",
                "ARKINDEX_API_SCHEMA_URL": "http://lul",
                "ARKINDEX_API_AUTH_SCHEME": "Roll",
                "ARKINDEX_API_TOKEN": "d6",
                "ARKINDEX_TASK_TOKEN": "d20",
                "ARKINDEX_API_CSRF_COOKIE": "thin-mint",
            },
            {
                "base_url": "http://lol",
                "schema_url": "http://lul",
                "auth_scheme": "Roll",
                "token": "d20",
                "csrf_cookie": "thin-mint",
            },
        ),
    ],
)
def test_options_from_env(mocker, env, expected):
    mocker.patch.dict(os.environ, env, clear=True)
    assert options_from_env() == expected



# -*- coding: utf-8 -*-
import re
from urllib.parse import urljoin

import typesystem
from typesystem.json_schema import JSONSchema

from arkindex.document import Document, Field, Link

SCHEMA_REF = typesystem.Object(
    properties={"$ref": typesystem.String(pattern="^#/components/schemas/")}
)
REQUESTBODY_REF = typesystem.Object(
    properties={"$ref": typesystem.String(pattern="^#/components/requestBodies/")}
)

definitions = typesystem.Definitions()

OPEN_API = typesystem.Schema(
    fields={
        "openapi": typesystem.String(),
        "servers": typesystem.Array(
            items=typesystem.Reference("Server", definitions=definitions),
            default=[],
        ),
        "paths": typesystem.Reference("Paths", definitions=definitions),
        "components": typesystem.Reference(
            "Components", default=None, definitions=definitions
        ),
    },
)

definitions["Server"] = typesystem.Schema(
    fields={
        "url": typesystem.String(),
    },
)

definitions["Paths"] = typesystem.Object(
    pattern_properties={
        "^/": typesystem.Reference("Path", definitions=definitions),
        "^x-": typesystem.Any(),
    },
    additional_properties=False,
)

definitions["Path"] = typesystem.Object(
    properties={
        "summary": typesystem.String(allow_blank=True),
        "get": typesystem.Reference("Operation", definitions=definitions),
        "put": typesystem.Reference("Operation", definitions=definitions),
        "post": typesystem.Reference("Operation", definitions=definitions),
        "delete": typesystem.Reference("Operation", definitions=definitions),
        "options": typesystem.Reference("Operation", definitions=definitions),
        "head": typesystem.Reference("Operation", definitions=definitions),
        "patch": typesystem.Reference("Operation", definitions=definitions),
        "trace": typesystem.Reference("Operation", definitions=definitions),
        "servers": typesystem.Array(
            items=typesystem.Reference("Server", definitions=definitions)
        ),
        "parameters": typesystem.Array(
            items=typesystem.Reference("Parameter", definitions=definitions)
        ),
        # TODO: | ReferenceObject
    },
    additional_properties=True,
)

definitions["Operation"] = typesystem.Object(
    properties={
        "operationId": typesystem.String(),
        "parameters": typesystem.Array(
            items=typesystem.Reference("Parameter", definitions=definitions)
        ),  # TODO: | ReferenceObject
        "requestBody": REQUESTBODY_REF
        | typesystem.Reference(
            "RequestBody", definitions=definitions
        ),  # TODO: RequestBody | ReferenceObject
        "deprecated": typesystem.Boolean(),
        "servers": typesystem.Array(
            items=typesystem.Reference("Server", definitions=definitions)
        ),
        # A custom property added by Arkindex to provide a hint to the API client
        # that this endpoint is paginated, which changes the behavior of ArkindexClient.paginate.
        "x-paginated": typesystem.Boolean(default=None),
    },
    additional_properties=True,
)

definitions["Parameter"] = typesystem.Schema(
    fields={
        "name": typesystem.String(),
        "in": typesystem.Choice(choices=["query", "header", "path", "cookie"]),
        "required": typesystem.Boolean(default=False),
        "deprecated": typesystem.Boolean(default=False),
        "schema": JSONSchema | SCHEMA_REF,
    },
)

definitions["RequestBody"] = typesystem.Object(
    properties={
        "content": typesystem.Object(
            additional_properties=typesystem.Reference(
                "MediaType", definitions=definitions
            )
        ),
        "required": typesystem.Boolean(),
    },
    additional_properties=True,
)

definitions["MediaType"] = typesystem.Object(
    properties={
        "schema": JSONSchema | SCHEMA_REF,
    },
    additional_properties=True,
)

definitions["Components"] = typesystem.Object(
    properties={
        "schemas": typesystem.Object(additional_properties=JSONSchema),
        "parameters": typesystem.Object(
            additional_properties=typesystem.Reference(
                "Parameter", definitions=definitions
            )
        ),
        "requestBodies": typesystem.Object(
            additional_properties=typesystem.Reference(
                "RequestBody", definitions=definitions
            )
        ),
    },
    additional_properties=True,
)

METHODS = ["get", "put", "post", "delete", "options", "head", "patch", "trace"]


def lookup(value, keys, default=None):
    for key in keys:
        try:
            value = value[key]
        except (KeyError, IndexError, TypeError):
            return default
    return value


def _simple_slugify(text):
    if text is None:
        return None
    text = text.lower()
    text = re.sub(r"[^a-z0-9]+", "_", text)
    text = re.sub(r"[_]+", "_", text)
    return text.strip("_")


class OpenAPI:
    def load(self, data):
        schema_definitions = self.get_schema_definitions(data)
        links = self.get_links(data, schema_definitions)

        return Document(links=links)

    def get_schema_definitions(self, data):
        definitions = typesystem.Definitions()
        schemas = lookup(data, ["components", "schemas"], {})
        for key, value in schemas.items():
            ref = f"#/components/schemas/{key}"
            definitions[ref] = typesystem.from_json_schema(
                value, definitions=definitions
            )
        return definitions

    def get_links(self, data, schema_definitions):
        """
        Return all the links in the document, laid out by operationId.
        """
        links = []

        for path, path_info in data.get("paths", {}).items():
            operations = {key: path_info[key] for key in path_info if key in METHODS}
            for operation, operation_info in operations.items():
                link = self.get_link(
                    path,
                    path_info,
                    operation,
                    operation_info,
                    schema_definitions,
                )
                if link is None:
                    continue

                links.append(link)

        return links

    def get_link(self, path, path_info, operation, operation_info, schema_definitions):
        """
        Return a single link in the document.
        """
        name = operation_info["operationId"]
        deprecated = operation_info.get("deprecated", False)
        paginated = operation_info.get("x-paginated")

        # Allow path info and operation info to override the base URL, which is normally managed by the ArkindexClient.
        base_url = lookup(path_info, ["servers", 0, "url"], default=None)
        base_url = lookup(operation_info, ["servers", 0, "url"], default=base_url)

        # Parameters are taken both from the path info, and from the operation.
        parameters = path_info.get("parameters", [])
        parameters += operation_info.get("parameters", [])

        fields = [
            self.get_field(parameter, schema_definitions) for parameter in parameters
        ]

        # TODO: Handle media type generically here...
        body_schema = lookup(
            operation_info, ["requestBody", "content", "application/json", "schema"]
        )

        encoding = None
        if body_schema:
            encoding = "application/json"
            if "$ref" in body_schema:
                ref = body_schema["$ref"]
                schema = schema_definitions.get(ref)
                prefix_length = len("#/components/schemas/")
                field_name = ref[prefix_length:].lower()
            else:
                schema = typesystem.from_json_schema(
                    body_schema, definitions=schema_definitions
                )
                field_name = "body"
            field_name = lookup(
                operation_info, ["requestBody", "x-name"], default=field_name
            )
            fields += [Field(name=field_name, location="body", schema=schema)]

        return Link(
            name=name,
            url=urljoin(base_url, path),
            method=operation,
            fields=fields,
            encoding=encoding,
            deprecated=deprecated,
            paginated=paginated,
        )

    def get_field(self, parameter, schema_definitions):
        """
        Return a single field in a link.
        """
        name = parameter.get("name")
        location = parameter.get("in")
        required = parameter.get("required", False)
        schema = parameter.get("schema")
        deprecated = parameter.get("deprecated", False)

        if schema is not None:
            if "$ref" in schema:
                ref = schema["$ref"]
                schema = schema_definitions.get(ref)
            else:
                schema = typesystem.from_json_schema(
                    schema, definitions=schema_definitions
                )

        return Field(
            name=name,
            location=location,
            required=required,
            schema=schema,
            deprecated=deprecated,
        )
